#include <stdio.h>
#include <stdlib.h>

typedef struct
{
	int value;
	short possible[9]; // possible[x-1] = 1 if the value x is possible for this field, otherwise possible[x-1] = 0
} sudoku_field;

typedef sudoku_field sudoku_net[9][9];

/**
 * Updates all fields in the row so the value is not possible for them anymore.
 * Returns list marking fields where the possibility info was updated.
 */
static short* disable_net_row(sudoku_net net, int col, int row, int value)
{
	int i;
	
	short* saved_row = (short*) malloc(sizeof(short) * 9);
	
	for (i = 1; i <= 9; i++)
	{
		if (i != col) 
		{
			if (net[i-1][row-1].possible[value-1])
			{
				net[i-1][row-1].possible[value-1] = 0;
				saved_row[i-1] = 1;
			}
			else saved_row[i-1] = 0;
		}
	}

	return saved_row;
}

/**
 * Restores possibility info for fields in the given row.
 */
static void enable_net_row(sudoku_net net, int col, int row, int value, short* saved_row)
{
	int i;

	for (i = 1; i <= 9; i++)
	{
		if (i != col) 
		{
			if (saved_row[i-1])
			{
				net[i-1][row-1].possible[value-1] = saved_row[i-1];
			}
		}
	}

	free(saved_row);
}

/**
 * Updates all fields in the column so the value is not possible for them anymore.
 * Returns list marking fields where the possibility info was updated.
 */
static short* disable_net_col(sudoku_net net, int col, int row, int value)
{
	int i;
	
	short* saved_col = (short*) malloc(sizeof(short) * 9);
	
	for (i = 1; i <= 9; i++)
	{
		if (i != row) 
		{
			if (net[col-1][i-1].possible[value-1])
			{
				net[col-1][i-1].possible[value-1] = 0;
				saved_col[i-1] = 1;
			}
			else saved_col[i-1] = 0;
		}
	}

	return saved_col;
}

/**
 * Restores possibility info for fields in the given column.
 */
static void enable_net_col(sudoku_net net, int col, int row, int value, short* saved_col)
{
	int i;

	for (i = 1; i <= 9; i++)
	{
		if (i != row) 
		{
			if (saved_col[i-1])
			{
				net[col-1][i-1].possible[value-1] = saved_col[i-1];
			}
		}
	}

	free(saved_col);
}

/**
 * Updates all fields in the 3x3 subnet so the value is not possible for them anymore.
 * Returns list marking fields where the possibility info was updated.
 */
static short* disable_net_3x3(sudoku_net net, int col, int row, int value)
{
	int i,j;
	
	int square_col_off = (int) ((col-1) / 3);
	int square_row_off = (int) ((row-1) / 3);

	short* saved_3x3 = (short*) malloc(sizeof(short) * 9);

	for (i = 0; i < 3; i++)
	{
		for (j = 0; j < 3; j++)
		{
			int cur_3x3_col = square_col_off*3 + 1 + i;
			int cur_3x3_row = square_row_off*3 + 1 + j;
					
			if ((cur_3x3_col != col) || (cur_3x3_row != row))
			{
				if (net[cur_3x3_col-1][cur_3x3_row-1].possible[value-1])
				{
					net[cur_3x3_col-1][cur_3x3_row-1].possible[value-1] = 0;
					saved_3x3[i*3+j] = 1;
				}
				else saved_3x3[i*3+j] = 0;
			}
		}
	}

	return saved_3x3;
}

/**
 * Restores possibility info for fields in the 3x3 subnet.
 */
static void enable_net_3x3(sudoku_net net, int col, int row, int value, short* saved_3x3)
{
	int i,j;

	int square_col_off = (int) ((col-1) / 3);
	int square_row_off = (int) ((row-1) / 3);

	for (i = 0; i < 3; i++)
	{
		for (j = 0; j < 3; j++)
		{
			int cur_3x3_col = square_col_off*3 + 1 + i;
			int cur_3x3_row = square_row_off*3 + 1 + j;
			
			if ((cur_3x3_col != col) || (cur_3x3_row != row))
			{
				if (saved_3x3[i*3+j])
				{
					net[cur_3x3_col-1][cur_3x3_row-1].possible[value-1] = saved_3x3[i*3+j];
				}
			}
		}
	}

	free(saved_3x3);
}

static void init_field(sudoku_net net, int col, int row, int value)
{
	int i;
	
	net[col-1][row-1].value = value;
	for (i = 1; i <= 9; i++) net[col-1][row-1].possible[i-1] = (i == value ? 1 : 0);

	free(disable_net_row(net, col, row, value));
	free(disable_net_col(net, col, row, value));

	free(disable_net_3x3(net, col, row, value));
}

static void init_net(sudoku_net net)
{
	int i,j,k;
	
	for (i = 1; i <= 9; i++)
	{
		for (j = 1; j <= 9; j++)
		{
			net[i-1][j-1].value = -1;
			for (k = 1; k <= 9; k++) net[i-1][j-1].possible[k-1] = 1;
		}
	}
}

static int load_net(sudoku_net net, char* file_name)
{
	int col, row;

	FILE* f = fopen(file_name, "r");

	char* line = (char*) malloc(sizeof(char)*12);

	for (row = 1; row <= 9; row++)
	{
		char* s = fgets(line, 11, f);
		if (s == NULL) return 0; // error (premature EOF)

		for (col = 1; col <= 9; col++)
		{
			if ((line[col-1] != ' ') && ((line[col-1] < '1') || (line[col-1] > '9'))) return 0; // error (invalid character)

			if (line[col-1] != ' ') init_field(net, col, row, line[col-1]-'0');
		}
	}
	
	free(line);

	fclose(f);
	
	return 1;
}

static void print_net(sudoku_net net, const char* title)
{
	int x,y;
	
	printf("%s\n", title);
	
	// init sudoku net
	for (y = 1; y <= 9; y++)
	{
		for (x = 1; x <= 9; x++)
		{
			if (net[x-1][y-1].value == -1) printf(" ");
			else printf("%d", net[x-1][y-1].value);
		}
		printf("\n");
	}
}

static int solve_net(sudoku_net net, int col, int row)
{
	if ((col > 9) || (row > 9)) return 1; // OK, we have the solution
	
	int cur_value = 1;
	while (cur_value <= 9)
	{
		if (net[col-1][row-1].possible[cur_value-1])
		{
			// OK, it is possible to put this value to the field

			net[col-1][row-1].value = cur_value;

			short* saved_possible_row;
			short* saved_possible_col;
			short* saved_possible_3x3;

			// update possible values for fields in the same row and column
			saved_possible_row = disable_net_row(net, col, row, cur_value);
			saved_possible_col = disable_net_col(net, col, row, cur_value);
		
			// update possible values for fields in the same 3x3 square
			saved_possible_3x3 = disable_net_3x3(net, col, row, cur_value);

			int next_col = col+1;
			int next_row = row;
			if (next_col > 9)
			{
				next_col = 1;
				next_row++;
			}

			int result = solve_net(net, next_col, next_row);
			
			if (result)
			{
				// the rest of the sudoku net is successfully solved
				return 1;
			}
			else
			{
				// failed to solve the rest of the sudoku net with assigned cur_value, so we must backtrack and try another possible value
				
				net[col-1][row-1].value = -1;

				// restore possible values for fields in the same row and column
				enable_net_row(net, col, row, cur_value, saved_possible_row);
				enable_net_col(net, col, row, cur_value, saved_possible_col);

				// restore possible values for fields in the same 3x3 square
				enable_net_3x3(net, col, row, cur_value, saved_possible_3x3);
			}
		}

		cur_value++;
	}

	return 0; // no number is applicable for the current field
}


int main(int argc, char** argv)
{
	if (argc < 2) 
	{
		printf("Usage: sudoku <input file>\n");
		return 1;
	}
	
	sudoku_net net;

	init_net(net);

	if (!load_net(net, argv[1]))
	{
		printf("Failed to load the input sudoku net\n");
		return 1;
	}
	
	print_net(net, "\nPuzzle");
	
	int result = solve_net(net, 1, 1);

	if (result) print_net(net, "\nSolution");
	else printf("\nFailed to solve the given sudoku net\n");

	return 0;
}
